import re
import subprocess  # nosec
from pathlib import Path
from typing import Optional

from . import _contents

_PYTHON_VERSION_REGEX = re.compile(r"^3\.\d+\.\d+$")


def create(
    package: Path,
    author: str,
    author_email: str,
    python_version: str,
    description: Optional[str] = None,
    init_git: bool = True,
    tests_dir: bool = False,
) -> None:
    if _PYTHON_VERSION_REGEX.match(python_version) is None:
        raise ValueError(f"invalid Python version '{python_version}' - must of the form 3.X.X")
    if description is None:
        description = ""

    srcdir = package.name.replace("-", "_")

    package.mkdir(parents=True, exist_ok=True)
    package.joinpath("src", srcdir).mkdir(parents=True, exist_ok=True)
    package.joinpath("requirements").mkdir(exist_ok=True)
    with package.joinpath(".flake8").open("w") as f:
        f.write(_contents.FLAKE8)
    with package.joinpath(".gitignore").open("w") as f:
        f.write(_contents.GITIGNORE)
    with package.joinpath(".pre-commit-config.yaml").open("w") as f:
        f.write(_contents.PRECOMMIT.format(package_name=package.name))
    with package.joinpath("pyproject.toml").open("w") as f:
        target_version = f"py{''.join(python_version.split('.')[:2])}"
        f.write(_contents.PYPROJECT.format(package_name=package.name, target_version=target_version))
    with package.joinpath("README.md").open("w") as f:
        f.write(_contents.README.format(package_name=package.name))
    with package.joinpath("setup.py").open("w") as f:
        f.write(
            _contents.SETUP.format(
                package_name=package.name,
                author=author,
                author_email=author_email,
                description=description,
                python_version=python_version,
            )
        )
    package.joinpath("requirements", "requirements.in").touch(exist_ok=True)
    with package.joinpath("requirements", "dev-requirements.in").open("w") as f:
        f.write(_contents.DEVREQUIREMENTS)
    package.joinpath("src", srcdir, "__init__.py").touch(exist_ok=True)
    if init_git:
        subprocess.run(["git", "init"], cwd=package)  # nosec
    if tests_dir:
        package.joinpath("tests").mkdir(exist_ok=True)


__all__ = ["create"]



from . import _contents


def test_pyproject():
    expected = """[build-system]
requires = ["setuptools>=42", "wheel"]
build-backend = "setuptools.build_meta"

[tool.black]
line-length = 120
target-version = ["py36"]

[tool.isort]
profile = "black"
"""
    actual = _contents.PYPROJECT.format(package_name="foo", target_version="py36")
    assert actual == expected


def test_readme():
    expected = """# foo
"""
    actual = _contents.README.format(package_name="foo")
    assert actual == expected


def test_setup():
    expected = """from pathlib import Path

from setuptools import find_packages, setup

HERE = Path(__file__).parent
README = HERE.joinpath("README.md").read_text()
REQUIREMENTS = HERE.joinpath("requirements", "requirements.in").read_text().split()

setup(
    name="foo",
    author="bar",
    author_email="baz@cool.com",
    description="hi there",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=REQUIREMENTS,
    python_requires=">=3.6.0",
)
"""
    actual = _contents.SETUP.format(
        package_name="foo", author="bar", author_email="baz@cool.com", description="hi there", python_version="3.6.0"
    )
    assert actual == expected


def test_precommit():
    expected = """# See https://pre-commit.com for more information
# See https://pre-commit.com/hooks.html for more hooks
repos:
  - repo: https://github.com/pre-commit/pre-commit-hooks
    rev: v4.1.0
    hooks:
      - id: trailing-whitespace
      - id: end-of-file-fixer
      - id: check-json
      - id: check-yaml
      - id: check-added-large-files
  - repo: https://github.com/PyCQA/isort
    rev: 5.10.1
    hooks:
      - id: isort
  - repo: https://github.com/psf/black
    rev: 22.1.0
    hooks:
      - id: black
        language_version: python3
  - repo: https://gitlab.com/pycqa/flake8
    rev: 4.0.1
    hooks:
      - id: flake8
        additional_dependencies: [flake8-bugbear]
  - repo: https://github.com/PyCQA/bandit
    rev: 1.7.1
    hooks:
      - id: bandit
        args: ["-x", "*/**/*_test.py"]
  - repo: local
    hooks:
      - id: pytest
        name: pytest
        entry: pytest src/foo
        language: system
        pass_filenames: false
        types: [python]
"""
    actual = _contents.PRECOMMIT.format(package_name="foo")
    assert actual == expected



import json
from pathlib import Path
from typing import Dict

from appdirs import user_config_dir


def load_custom_config(path: Path) -> Dict[str, str]:
    config = json.loads(path.read_text())
    if not isinstance(config, dict):
        raise ValueError("the config object must be a flat dictionary")
    return config


def try_load_default_config() -> Dict[str, str]:
    config_path = Path(user_config_dir("pypk", "Pyrite AI")).joinpath("config.json")
    try:
        config = json.loads(config_path.read_text())
    except FileNotFoundError:
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with config_path.open("w") as f:
            json.dump({}, f)
        return {}
    except (OSError, json.decoder.JSONDecodeError):
        print("[Warning] failed to read default config file: invalid JSON")
        return {}
    else:
        return config


def save_as_default_config(cfg: Dict[str, str]):
    config_path = Path(user_config_dir("pypk", "Pyrite AI")).joinpath("config.json")
    with config_path.open("w") as f:
        json.dump(cfg, f)


__all__ = ["load_custom_config", "save_as_default_config", "try_load_default_config"]



FLAKE8 = """[flake8]
max-line-length = 120
extend-ignore = E203
max-complexity = 12
select = B,C,E,F,W,B9
"""
PYPROJECT = """[build-system]
requires = ["setuptools>=42", "wheel"]
build-backend = "setuptools.build_meta"

[tool.black]
line-length = 120
target-version = ["{target_version}"]

[tool.isort]
profile = "black"
"""
README = """# {package_name}
"""
SETUP = """from pathlib import Path

from setuptools import find_packages, setup

HERE = Path(__file__).parent
README = HERE.joinpath("README.md").read_text()
REQUIREMENTS = HERE.joinpath("requirements", "requirements.in").read_text().split()

setup(
    name="{package_name}",
    author="{author}",
    author_email="{author_email}",
    description="{description}",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={{"": "src"}},
    install_requires=REQUIREMENTS,
    python_requires=">={python_version}",
)
"""
PRECOMMIT = """# See https://pre-commit.com for more information
# See https://pre-commit.com/hooks.html for more hooks
repos:
  - repo: https://github.com/pre-commit/pre-commit-hooks
    rev: v4.1.0
    hooks:
      - id: trailing-whitespace
      - id: end-of-file-fixer
      - id: check-json
      - id: check-yaml
      - id: check-added-large-files
  - repo: https://github.com/PyCQA/isort
    rev: 5.10.1
    hooks:
      - id: isort
  - repo: https://github.com/psf/black
    rev: 22.1.0
    hooks:
      - id: black
        language_version: python3
  - repo: https://gitlab.com/pycqa/flake8
    rev: 4.0.1
    hooks:
      - id: flake8
        additional_dependencies: [flake8-bugbear]
  - repo: https://github.com/PyCQA/bandit
    rev: 1.7.1
    hooks:
      - id: bandit
        args: ["-x", "*/**/*_test.py"]
  - repo: local
    hooks:
      - id: pytest
        name: pytest
        entry: pytest src/{package_name}
        language: system
        pass_filenames: false
        types: [python]
"""
DEVREQUIREMENTS = """bandit
black
flake8
flake8-bugbear
isort
pre-commit
pytest
safety
"""
GITIGNORE = """# Created by https://www.toptal.com/developers/gitignore/api/macos,windows,linux,python,vim,emacs,sublimetext,visualstudiocode,pycharm
# Edit at https://www.toptal.com/developers/gitignore?templates=macos,windows,linux,python,vim,emacs,sublimetext,visualstudiocode,pycharm

### Emacs ###
# -*- mode: gitignore; -*-
*~
\\#*\\#
/.emacs.desktop
/.emacs.desktop.lock
*.elc
auto-save-list
tramp
.\\#*

# Org-mode
.org-id-locations
*_archive

# flymake-mode
*_flymake.*

# eshell files
/eshell/history
/eshell/lastdir

# elpa packages
/elpa/

# reftex files
*.rel

# AUCTeX auto folder
/auto/

# cask packages
.cask/
dist/

# Flycheck
flycheck_*.el

# server auth directory
/server/

# projectiles files
.projectile

# directory configuration
.dir-locals.el

# network security
/network-security.data


### Linux ###

# temporary files which can be created if a process still has a handle open of a deleted file
.fuse_hidden*

# KDE directory preferences
.directory

# Linux trash folder which might appear on any partition or disk
.Trash-*

# .nfs files are created when an open file is removed but is still being accessed
.nfs*

### macOS ###
# General
.DS_Store
.AppleDouble
.LSOverride

# Icon must end with two \r
Icon


# Thumbnails
._*

# Files that might appear in the root of a volume
.DocumentRevisions-V100
.fseventsd
.Spotlight-V100
.TemporaryItems
.Trashes
.VolumeIcon.icns
.com.apple.timemachine.donotpresent

# Directories potentially created on remote AFP share
.AppleDB
.AppleDesktop
Network Trash Folder
Temporary Items
.apdisk

### PyCharm ###
# Covers JetBrains IDEs: IntelliJ, RubyMine, PhpStorm, AppCode, PyCharm, CLion, Android Studio, WebStorm and Rider
# Reference: https://intellij-support.jetbrains.com/hc/en-us/articles/206544839

# User-specific stuff
.idea/**/workspace.xml
.idea/**/tasks.xml
.idea/**/usage.statistics.xml
.idea/**/dictionaries
.idea/**/shelf

# AWS User-specific
.idea/**/aws.xml

# Generated files
.idea/**/contentModel.xml

# Sensitive or high-churn files
.idea/**/dataSources/
.idea/**/dataSources.ids
.idea/**/dataSources.local.xml
.idea/**/sqlDataSources.xml
.idea/**/dynamic.xml
.idea/**/uiDesigner.xml
.idea/**/dbnavigator.xml

# Gradle
.idea/**/gradle.xml
.idea/**/libraries

# Gradle and Maven with auto-import
# When using Gradle or Maven with auto-import, you should exclude module files,
# since they will be recreated, and may cause churn.  Uncomment if using
# auto-import.
# .idea/artifacts
# .idea/compiler.xml
# .idea/jarRepositories.xml
# .idea/modules.xml
# .idea/*.iml
# .idea/modules
# *.iml
# *.ipr

# CMake
cmake-build-*/

# Mongo Explorer plugin
.idea/**/mongoSettings.xml

# File-based project format
*.iws

# IntelliJ
out/

# mpeltonen/sbt-idea plugin
.idea_modules/

# JIRA plugin
atlassian-ide-plugin.xml

# Cursive Clojure plugin
.idea/replstate.xml

# SonarLint plugin
.idea/sonarlint/

# Crashlytics plugin (for Android Studio and IntelliJ)
com_crashlytics_export_strings.xml
crashlytics.properties
crashlytics-build.properties
fabric.properties

# Editor-based Rest Client
.idea/httpRequests

# Android studio 3.1+ serialized cache file
.idea/caches/build_file_checksums.ser

### PyCharm Patch ###
# Comment Reason: https://github.com/joeblau/gitignore.io/issues/186#issuecomment-215987721

# *.iml
# modules.xml
# .idea/misc.xml
# *.ipr

# Sonarlint plugin
# https://plugins.jetbrains.com/plugin/7973-sonarlint
.idea/**/sonarlint/

# SonarQube Plugin
# https://plugins.jetbrains.com/plugin/7238-sonarqube-community-plugin
.idea/**/sonarIssues.xml

# Markdown Navigator plugin
# https://plugins.jetbrains.com/plugin/7896-markdown-navigator-enhanced
.idea/**/markdown-navigator.xml
.idea/**/markdown-navigator-enh.xml
.idea/**/markdown-navigator/

# Cache file creation bug
# See https://youtrack.jetbrains.com/issue/JBR-2257
.idea/$CACHE_FILE$

# CodeStream plugin
# https://plugins.jetbrains.com/plugin/12206-codestream
.idea/codestream.xml

### Python ###
# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions
*.so

# Distribution / packaging
.Python
build/
develop-eggs/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/
cover/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py
db.sqlite3
db.sqlite3-journal

# Flask stuff:
instance/
.webassets-cache

# Scrapy stuff:
.scrapy

# Sphinx documentation
docs/_build/

# PyBuilder
.pybuilder/
target/

# Jupyter Notebook
.ipynb_checkpoints

# IPython
profile_default/
ipython_config.py

# pyenv
#   For a library or package, you might want to ignore these files since the code is
#   intended to run in multiple environments; otherwise, check them in:
# .python-version

# pipenv
#   According to pypa/pipenv#598, it is recommended to include Pipfile.lock in version control.
#   However, in case of collaboration, if having platform-specific dependencies or dependencies
#   having no cross-platform support, pipenv may install dependencies that don't work, or not
#   install all needed dependencies.
#Pipfile.lock

# poetry
#   Similar to Pipfile.lock, it is generally recommended to include poetry.lock in version control.
#   This is especially recommended for binary packages to ensure reproducibility, and is more
#   commonly ignored for libraries.
#   https://python-poetry.org/docs/basic-usage/#commit-your-poetrylock-file-to-version-control
#poetry.lock

# PEP 582; used by e.g. github.com/David-OConnor/pyflow
__pypackages__/

# Celery stuff
celerybeat-schedule
celerybeat.pid

# SageMath parsed files
*.sage.py

# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/

# Spyder project settings
.spyderproject
.spyproject

# Rope project settings
.ropeproject

# mkdocs documentation
/site

# mypy
.mypy_cache/
.dmypy.json
dmypy.json

# Pyre type checker
.pyre/

# pytype static type analyzer
.pytype/

# Cython debug symbols
cython_debug/

# PyCharm
#  JetBrains specific template is maintainted in a separate JetBrains.gitignore that can
#  be found at https://github.com/github/gitignore/blob/main/Global/JetBrains.gitignore
#  and can be added to the global gitignore or merged into this file.  For a more nuclear
#  option (not recommended) you can uncomment the following to ignore the entire idea folder.
#.idea/

### SublimeText ###
# Cache files for Sublime Text
*.tmlanguage.cache
*.tmPreferences.cache
*.stTheme.cache

# Workspace files are user-specific
*.sublime-workspace

# Project files should be checked into the repository, unless a significant
# proportion of contributors will probably not be using Sublime Text
# *.sublime-project

# SFTP configuration file
sftp-config.json
sftp-config-alt*.json

# Package control specific files
Package Control.last-run
Package Control.ca-list
Package Control.ca-bundle
Package Control.system-ca-bundle
Package Control.cache/
Package Control.ca-certs/
Package Control.merged-ca-bundle
Package Control.user-ca-bundle
oscrypto-ca-bundle.crt
bh_unicode_properties.cache

# Sublime-github package stores a github token in this file
# https://packagecontrol.io/packages/sublime-github
GitHub.sublime-settings

### Vim ###
# Swap
[._]*.s[a-v][a-z]
!*.svg  # comment out if you don't need vector files
[._]*.sw[a-p]
[._]s[a-rt-v][a-z]
[._]ss[a-gi-z]
[._]sw[a-p]

# Session
Session.vim
Sessionx.vim

# Temporary
.netrwhist
# Auto-generated tag files
tags
# Persistent undo
[._]*.un~

### VisualStudioCode ###
.vscode/*
!.vscode/settings.json
!.vscode/tasks.json
!.vscode/launch.json
!.vscode/extensions.json
!.vscode/*.code-snippets

# Local History for Visual Studio Code
.history/

# Built Visual Studio Code Extensions
*.vsix

### VisualStudioCode Patch ###
# Ignore all local history of files
.history
.ionide

# Support for Project snippet scope

### Windows ###
# Windows thumbnail cache files
Thumbs.db
Thumbs.db:encryptable
ehthumbs.db
ehthumbs_vista.db

# Dump file
*.stackdump

# Folder config file
[Dd]esktop.ini

# Recycle Bin used on file shares
$RECYCLE.BIN/

# Windows Installer files
*.cab
*.msi
*.msix
*.msm
*.msp

# Windows shortcuts
*.lnk

# End of https://www.toptal.com/developers/gitignore/api/macos,windows,linux,python,vim,emacs,sublimetext,visualstudiocode,pycharm
"""



from pathlib import Path

from setuptools import find_packages, setup

HERE = Path(__file__).parent
README = HERE.joinpath("README.md").read_text()
REQUIREMENTS = HERE.joinpath("requirements", "requirements.in").read_text().split()

setup(
    name="pypk",
    author="Miller Wilt",
    author_email="miller@pyriteai.com",
    description="An opinionated Python package template generator",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/PyriteAI/pypk",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages("src"),
    package_dir={"": "src"},
    use_scm_version={"write_to": "src/pypk/version.py"},
    entry_points={"console_scripts": ["pypk=pypk.main:app"]},
    setup_requires=["setuptools_scm"],
    install_requires=REQUIREMENTS,
    python_requires=">=3.6.0",
)



import json
import sys
from pathlib import Path
from typing import NoReturn, Optional

import typer

from . import core
from .config import load_custom_config, save_as_default_config, try_load_default_config
from .version import version as pypk_version

_CURRENT_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}.0"

app = typer.Typer()


def exit_with_status(msg: str, code: int = 1) -> NoReturn:
    typer.echo(msg, err=True)
    sys.exit(code)


@app.command()
def create(
    package: Path = typer.Argument(..., help="Path to root package directory"),  # noqa: B008
    config: Path = typer.Option(None, "--config", "-c", help="Path to config file"),  # noqa: B008
    author: str = typer.Option(None, "--author", "-a", help="Author's name"),  # noqa: B008
    email: str = typer.Option(None, "--email", "-e", help="Author's email"),  # noqa: B008
    description: Optional[str] = typer.Option(  # noqa: B008
        None, "--description", "-d", help="Description to be placed in setup.py"
    ),
    python_version: str = typer.Option(  # noqa: B008
        _CURRENT_VERSION, "--py-version", "-p", help="Minimum Python version supported"
    ),
    init_git: bool = typer.Option(None, "--init-git/--skip-git-init", help="Toggle Git initialization"),  # noqa: B008
    tests_dir: bool = typer.Option(  # noqa: B008
        None, "--create-tests-dir/--skip-tests-dir", help="Create a top-level tests directory"
    ),
) -> None:
    try:
        if config is not None:
            cfg = load_custom_config(config)
        else:
            cfg = try_load_default_config()
    except Exception as e:
        exit_with_status(f"[Error] failed to load config: '{e}'")

    author = author if author else cfg.get("author")
    email = email if email else cfg.get("email")
    python_version = python_version if python_version else cfg.get("py_version")
    init_git = init_git if init_git is not None else cfg.get("init_git", False)
    tests_dir = tests_dir if tests_dir is not None else cfg.get("tests_dir", False)

    if author is None:
        exit_with_status("[Error] 'author' must be specified in either the config or via command line")
    if email is None:
        exit_with_status("[Error] 'email' must be specified in either the config or via command line")
    if version is None:
        exit_with_status("[Error] 'version' must be specified in either the config or via command line")
    if not isinstance(init_git, bool):
        exit_with_status("[Error] 'init_git' must be a boolean")
    if not isinstance(tests_dir, bool):
        exit_with_status("[Error] 'tests_dir' must be a boolean")

    core.create(package, author, email, python_version, description=description, init_git=init_git, tests_dir=tests_dir)


@app.command()
def config(key: str, value: Optional[str] = typer.Argument(None)):  # noqa: B008
    try:
        cfg = try_load_default_config()
    except Exception as e:
        exit_with_status(f"[Error] failed to load config: '{e}'")

    if key not in ["author", "email", "py_version", "init_git", "tests_dir"]:
        exit_with_status(f"[Error] invalid key '{key}'")

    if value:
        if key in ["init_git", "tests_dir"]:
            value = json.loads(value.lower())  # hacky, but it works
            if not isinstance(value, bool):
                exit_with_status(f"[Error] '{key}' must be a boolean")
        cfg[key] = value
        save_as_default_config(cfg)
    else:
        value = cfg.get(key)
        if value:
            typer.echo(f"'{key}' = '{value}'")
        else:
            typer.echo(f"'{key}' is not set")


@app.command()
def version():
    typer.echo(pypk_version)


__all__ = ["app"]
